# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/03_ityp.ipynb.

# %% auto 0
__all__ = ['IterStr', 'IterNum', 'IterBool']

# %% ../nbs/03_ityp.ipynb 4
from typing import (
    Type, TypeAlias, 
    ClassVar, Any, Optional, Callable, Iterable, 
)
from numbers import Number

# %% ../nbs/03_ityp.ipynb 5
from .atyp import Boolish, Numeric, StrLike, Module

# %% ../nbs/03_ityp.ipynb 7
IterStr: TypeAlias = Iterable[str]
IterNum: TypeAlias = Iterable[Numeric]
IterBool: TypeAlias = Iterable[Boolish]



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/atyp',
                'doc_host': 'https://dsm-72.github.io',
                'git_url': 'https://github.com/dsm-72/atyp',
                'lib_path': 'atyp'},
  'syms': { 'atyp.atyp': { 'atyp.atyp.ATyp': ('atyp.html#atyp', 'atyp/atyp.py'),
                           'atyp.atyp.Alias': ('atyp.html#alias', 'atyp/atyp.py'),
                           'atyp.atyp.Alias.__init_subclass__': ('atyp.html#alias.__init_subclass__', 'atyp/atyp.py'),
                           'atyp.atyp.AliasMeta': ('atyp.html#aliasmeta', 'atyp/atyp.py'),
                           'atyp.atyp.AliasMeta.__call__': ('atyp.html#aliasmeta.__call__', 'atyp/atyp.py'),
                           'atyp.atyp.AliasMeta.__instancecheck__': ('atyp.html#aliasmeta.__instancecheck__', 'atyp/atyp.py'),
                           'atyp.atyp.AliasMeta.__new__': ('atyp.html#aliasmeta.__new__', 'atyp/atyp.py'),
                           'atyp.atyp.AliasMeta.check': ('atyp.html#aliasmeta.check', 'atyp/atyp.py'),
                           'atyp.atyp.AliasMeta.default_kwds': ('atyp.html#aliasmeta.default_kwds', 'atyp/atyp.py'),
                           'atyp.atyp.AliasMeta.getclsname': ('atyp.html#aliasmeta.getclsname', 'atyp/atyp.py'),
                           'atyp.atyp.AliasMeta.getqualname': ('atyp.html#aliasmeta.getqualname', 'atyp/atyp.py'),
                           'atyp.atyp.AliasMeta.setattrs': ('atyp.html#aliasmeta.setattrs', 'atyp/atyp.py'),
                           'atyp.atyp.AliasMeta.settypes': ('atyp.html#aliasmeta.settypes', 'atyp/atyp.py'),
                           'atyp.atyp.aliascheck': ('atyp.html#aliascheck', 'atyp/atyp.py'),
                           'atyp.atyp.eqliteral': ('atyp.html#eqliteral', 'atyp/atyp.py'),
                           'atyp.atyp.isliteral': ('atyp.html#isliteral', 'atyp/atyp.py')},
            'atyp.biol': {},
            'atyp.bool': {'atyp.bool.Bool01': ('bool.html#bool01', 'atyp/bool.py')},
            'atyp.btyp': {},
            'atyp.byte': {},
            'atyp.clsv': {},
            'atyp.cstm': {},
            'atyp.ctyp': {},
            'atyp.data': {},
            'atyp.dtyp': {},
            'atyp.enum': {},
            'atyp.imgs': {},
            'atyp.imps': {},
            'atyp.insp': {},
            'atyp.iter': {},
            'atyp.ityp': {},
            'atyp.list': {},
            'atyp.logr': {},
            'atyp.mtyp': {},
            'atyp.npty': {},
            'atyp.nums': {},
            'atyp.path': {},
            'atyp.pdty': {'atyp.pdty.BoolIndex': ('pdty.html#boolindex', 'atyp/pdty.py')},
            'atyp.plot': {},
            'atyp.ptyp': {},
            'atyp.qtyp': {},
            'atyp.rich': {},
            'atyp.rtyp': {},
            'atyp.sets': {},
            'atyp.spty': {},
            'atyp.strs': {},
            'atyp.tens': {},
            'atyp.tgrd': {},
            'atyp.type': {},
            'atyp.typq': {},
            'atyp.typr': {},
            'atyp.ztyp': {}}}



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/libs/06_rtyp.ipynb.

# %% auto 0
__all__ = ['RichTree', 'RichText', 'RichCons', 'RichProg', 'TyperApp']

# %% ../nbs/libs/06_rtyp.ipynb 4
from typing import TypeAlias, ForwardRef

# %% ../nbs/libs/06_rtyp.ipynb 6
try:
    from rich.tree import Tree
    from rich.text import Text
    from rich.console import Console
    from rich.progress import Progress
    tree = Tree
    text = Text
    cons = Console
    prog = Progress
    
except ImportError:
    tree = ForwardRef('rich.tree.Tree', is_class=True)
    text = ForwardRef('rich.text.Text', is_class=True)
    cons = ForwardRef('rich.console.Console', is_class=True)
    prog = ForwardRef('rich.progress.Progress', is_class=True)

# %% ../nbs/libs/06_rtyp.ipynb 7
RichTree: TypeAlias = tree
RichText: TypeAlias = text
RichCons: TypeAlias = cons
RichProg: TypeAlias = prog

# %% ../nbs/libs/06_rtyp.ipynb 9
try:
    from typer import Typer
    tapp = Typer    
except ImportError:
    tapp = ForwardRef('typer.Typer', is_class=True)

# %% ../nbs/libs/06_rtyp.ipynb 10
TyperApp: TypeAlias = tapp



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/112_data.ipynb.

# %% auto 0
__all__ = []

# %% ../nbs/112_data.ipynb 6
#| export


# %% ../nbs/112_data.ipynb 8
from typing import TypeAlias, List, Union, Iterable, List

# %% ../nbs/112_data.ipynb 11
# try:
#     import numpy as np
#     arr = np.ndarray
#     mtx = np.matrix
#     dsr = np.DataSource
# except ImportError:
#     # Define a placeholder type if numpy is not installed
#     arr = 'np.ndarray'
#     mtx = 'np.matrix'
#     dsr = 'np.DataSource'

# %% ../nbs/112_data.ipynb 13
# try:
#     import pandas as pd
#     index = pd.Index
#     cindex = pd.CategoricalIndex
#     dframe = pd.DataFrame
#     series = pd.Series
# except ImportError:
#     # Define a placeholder type if pandas is not installed
#     index = 'pd.Index'    
#     dframe = 'pd.DataFrame'
#     series = 'pd.Series'
#     cindex = 'pd.CategoricalIndex'

# %% ../nbs/112_data.ipynb 15
# try:
#     import scipy as sp
#     try:
#         sparr = sp.sparse._arrays._sparray
#     except AttributeError:
#         sparr = sp.sparse.sparray
#     spmtx = sp.sparse.spmatrix
# except ImportError:
#     # Define a placeholder type if scipy is not installed
#     # sparr = 'sp.sparse.sparray'
#     # spmtx = 'sp.sparse.spmatrix'
#     sparr = 'scipy.sparse._base.sparray'
#     spmtx = 'scipy.sparse._matrix.spmatrix'

# %% ../nbs/112_data.ipynb 17
#| export


# %% ../nbs/112_data.ipynb 19
from .bool import Boolish, Bool01
from .nums import Numbers
from .tens import Tensor
from .npty import (NPArray, ndarray, NPMatrix, ndmatrix, DataSource)
from atyp.pdty import (
    Idx, Index, CIdx, CIndex, CatIndex, CategoricalIndex, 
    BoolIndex, BIdx, BooleanIndex,
    Series, DFrame, DataFrame, DataFrames,

    IndexLike, GroupKey
)
from .spty import (SPArray, SPMatrix)

# %% ../nbs/112_data.ipynb 22
NPArray, ndarray, NPMatrix, ndmatrix, DataSource
# NPArray: TypeAlias = arr
# ndarray: TypeAlias = arr
# NPMatrix: TypeAlias = mtx
# ndmatrix: TypeAlias = mtx
# DataSource: TypeAlias = dsr

# %% ../nbs/112_data.ipynb 24
Idx, Index, CIdx, CIndex, CatIndex, CategoricalIndex, BoolIndex, BIdx, BooleanIndex, Series, DFrame, DataFrame, DataFrames
# Idx: TypeAlias = index
# Index: TypeAlias = index


# CIdx: TypeAlias = cindex
# CIndex: TypeAlias = cindex
# CatIndex: TypeAlias = cindex
# CategoricalIndex: TypeAlias = cindex

# Series: TypeAlias = series

# DFrame: TypeAlias = dframe
# DataFrame: TypeAlias = dframe
# DataFrames: TypeAlias = List[DataFrame]

# %% ../nbs/112_data.ipynb 26
IndexLike
# IndexLike: TypeAlias = Union[Numbers, List[Boolish], Index]
# SeriesLike: TypeAlias = Union[list, Series, ndarray, Tensor, IndexLike]
# IterLike: TypeAlias = Union[Iterable, SeriesLike, IndexLike]

# %% ../nbs/112_data.ipynb 27
GroupKey
# GroupKey: TypeAlias = Union[str, int, float, tuple] # key for pandas groupby

# %% ../nbs/112_data.ipynb 29
SPArray, SPMatrix
# SPArray: TypeAlias = sparr
# SPMatrix: TypeAlias = spmtx



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_atyp.ipynb.

# %% auto 0
__all__ = ['isliteral', 'eqliteral', 'aliascheck', 'AliasMeta', 'Alias', 'ATyp']

# %% ../nbs/00_atyp.ipynb 6
import os, inspect
from functools import wraps

# %% ../nbs/00_atyp.ipynb 8
from typing import (
    Any, Type, Tuple, Union, Callable, Iterable, TypeGuard, Protocol, _ProtocolMeta,
    runtime_checkable, TypeVar, ParamSpec, TypeAlias, Self, Optional,
    Literal, get_args
)

# %% ../nbs/00_atyp.ipynb 10
#| export


# %% ../nbs/00_atyp.ipynb 12
from ntyp import notiterstr, alldtype, Types, Guard, guardtype

# %% ../nbs/00_atyp.ipynb 14
#| export

# %% ../nbs/00_atyp.ipynb 16
def isliteral(obj):
    return getattr(obj, '__name__', None) == 'Literal'
    
def eqliteral(val, obj):
    if not isliteral(obj): return False
    try: return val == get_args(obj)[0]            
    except: return False

# %% ../nbs/00_atyp.ipynb 17
def aliascheck(obj, types, attrs: dict = dict()):
    found = False
    for base in types:
        try:
            if isinstance(obj, base):
                found = True
                break
        except TypeError:
            try:
                if eqliteral(obj, base) or obj == base:
                    found = True
                    break       
            except: ...                

    if not found:
        return False        
    
    for attr, val in attrs.items():
        if attr == 'dtype' and notiterstr(obj) and alldtype(obj, val): continue                        
        if getattr(obj, attr, None) == val: continue
        return False
        raise ValueError(f'Expected {attr} to be {val}, got {getattr(obj, attr, None)}')
    return True

# %% ../nbs/00_atyp.ipynb 18
class AliasMeta(_ProtocolMeta):
    types: Types = tuple()
    attrs: dict = dict()

    @classmethod
    def default_kwds(cls, **kwargs) -> dict:
        kwds = dict(clsname = kwargs.get('clsname', True))
        return kwds

    def __new__(mcls, name, bases, dct, **kwargs):
        '''Create a new class instance.'''        
        kwds = mcls.default_kwds(**kwargs)
        dct['types'] = (types := kwargs.pop('types', tuple()))
        dct['attrs'] = (attrs := kwargs.pop('attrs', dict()))
        dct['check'] = classmethod(lambda cls, obj: aliascheck(obj, types, attrs))        
        new = super().__new__(mcls, name, bases, dct)
        new.__qualname__ = new.getqualname(*types, **kwds)
        new.settypes(*types, **kwds)
        return new

    def __call__(cls, *dtypes: Types, **kwargs):
        '''Allows using Not as a decorator with types to negate.'''        
        def decorator(kls):
            '''Decorator to apply negation to a class.'''
            clsname = kwargs.get('clsname', kls.__name__)
            bases = (kls, Alias) if Alias not in kls.__bases__ else (kls,)
            dct = dict(check = classmethod(lambda cls, obj: aliascheck(obj, cls.types, cls.attrs)))
            new = cls.__class__(clsname, bases, dct, types=dtypes, **kwargs)
            return new
        return decorator

    def __instancecheck__(self, ins) -> bool: 
        return self.check(ins)
    
    def check(self, obj) -> Guard:     
        return aliascheck(obj, self.types, self.attrs)
        return isinstance(obj, self.types)

    def getclsname(cls, clsname: Union[str, bool] = True) -> Union[str, bool]:
        if clsname == False: return clsname
        if inspect.isclass(clsname): return clsname.__name__
        elif isinstance(clsname, str): return clsname
        else: return cls.__name__
        
    def getqualname(cls, *types: Types, clsname: Union[str, bool] = True) -> str:
        tname = ', '.join([str(getattr(arg, '__name__', arg)) for arg in types or tuple()])
        qname = f'{{{tname}}}'        
        if (cname := cls.getclsname(clsname)): qname = f'{cname}({qname})'
        return qname
    
    def settypes(cls, *types: Types, clsname: str | bool = True,):
        qname = cls.getqualname(*types, clsname=clsname)
        setattr(cls, '__qualname__', qname)
        annots = getattr(getattr(cls, 'check', None), '__annotations__', {})
        annots['return'] = guardtype(*types)
        setattr(cls, 'types', types)

    def setattrs(cls, attrs: dict = dict()):        
        setattr(cls, 'attrs', attrs)

# %% ../nbs/00_atyp.ipynb 19
@runtime_checkable
class Alias(Protocol, metaclass=AliasMeta):
    @classmethod
    def __init_subclass__(cls, *args, types: Types = tuple(), attrs: dict = dict(), **kwargs) -> None:
        super().__init_subclass__()
        kwds = cls.default_kwds(**attrs)
        cls.settypes(*types, **kwds)
        cls.setattrs(attrs)

@wraps(Alias, updated=())
class ATyp: ...



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/30_ztyp.ipynb.

# %% auto 0
__all__ = ['FlagFunc', 'FlagFuncQ', 'PathFormatFunc', 'Mixin']

# %% ../nbs/30_ztyp.ipynb 4
import os
from typing import List, Union, TypeAlias, Callable, Optional, ParamSpec, Type, Callable, Tuple

# %% ../nbs/30_ztyp.ipynb 6
from .atyp import Boolish
from .qtyp import AnyQ, StrQ

# %% ../nbs/30_ztyp.ipynb 7
FlagFunc: TypeAlias = Callable[[AnyQ], Boolish]

FlagFuncQ: TypeAlias = Optional[FlagFunc]

# %% ../nbs/30_ztyp.ipynb 11
PathFormatFunc: TypeAlias = Callable[[os.PathLike, StrQ, StrQ, StrQ], str]

# %% ../nbs/30_ztyp.ipynb 13
Mixin: TypeAlias =  Union[Tuple[Type, Callable], 'MixinType']
